// ------ tshape.h
#ifndef TSHAPE_H
#define TSHAPE_H

#include <vector>
#include <algorithm>
#include <shape.h>

namespace GraphicsLibrary  {

struct Point {
   	int x, y;
    Point(int xp = 0, int yp = 0) : x(xp), y(yp) { }
	Point operator+(const Point& pt) const
		{ Point p(x,y); p.x += pt.x; p.y += pt.y; return p; }
	Point operator+=(const Point& pt)
		{ x += pt.x; y += pt.y; return *this; }
	Point operator-(const Point& pt) const
		{ Point p(x,y); p.x -= pt.x; p.y -= pt.y; return p; }
	Point operator-=(const Point& pt)
		{ x -= pt.x; y -= pt.y; return *this; }
	bool operator<(const Point& pt) const
		{ return (y == pt.y) ? (x < pt.x) : (y < pt.y); }
};

template <class T>
class TShape : public Shape {
protected:
    int x, y;		// position of shape
    T* tp;			// generic graphical rendering engine

    TShape(T* t = 0, int xp = 0, int yp = 0, Color clr = black) : 
			Shape(clr), x(xp), y(yp), tp(t)
		{  }
    TShape(T* t, Point pos, Color clr) : 
			Shape(clr), x(pos.x), y(pos.y), tp(t)
		{  }
	virtual void draw() = 0;
	virtual void drawpoint(int xp, int yp);
public:
	Point getposition() const
		{ return Point(x, y); }
};

template <class T>
void TShape<T>::drawpoint(int xp, int yp)
{
	tp->drawpoint(xp + x, yp + y, color);
}

} // namespace GraphicsLibrary

#endif
