// ------- line.h

#ifndef LINE_H
#define LINE_H
#include <graphics.h>

namespace GraphicsLibrary  {

template <class T>
class Line : public TShape<T>  {
    int ex, ey;   // end point of line
public:
	// --- default ctor for std::vector
	Line() : ex(0), ey(0)
		{  }
    Line(T& t, int x1, int y1, int x2, int y2, Color clr = black) : 
                TShape<T>(&t, x1, y1, clr), ex(x2), ey(y2)
	    {  }
    Line(T& t, Point beg, Point end, Color clr = black) : 
				TShape<T>(&t, beg, clr), ex(end.x), ey(end.y)
    	{  }
    virtual void draw();
};

template <class T>
void Line<T>::draw()
{
	int begx = 0;		// beginning x coordinate
	int endx = ex - x;	// beginning y coordinate
	int begy = 0;		// ending x coordinate
	int endy = ey - y;	// ending y coordinate
    if (begx == endx)  {
        // --- vertical line
        if (endy < begy)
            std::swap(begy, endy);
		while (begy <= endy)
			drawpoint(begx, begy++);
    }
    else if (begy == endy)  {
        // ---- horizontal line
        if (endx < begx)
            std::swap(begx, endx);
		while (begx <= endx)
			drawpoint(begx++, begy);
    }
	else	{
		// --- force left to right plotting
		if (begx > endx)	{
			std::swap(begx, endx);
			std::swap(begy, endy);
		}
		// --- build the first point
		drawpoint(begx, begy);
		int lenx = endx-begx;	// length of the x axes
		int leny = endy-begy;	// length of the y axes
		int inctrack = 1;		// increment the axes tracked
		int incslope = 1;		// increment the slope axes
		int str, len;
		int* slope;
		int* track;
		if (leny < 0)	{
			// -- the y axes goes in reverse
			leny = -leny;
			if (leny < lenx)
				incslope = -1;
			else
				inctrack = -1;
		}
		if (leny < lenx)	{
			// --- track the x axes, slope the y axes
			str = leny * 2;
			len = lenx;
			slope = &begy;
			track = &begx;
		}
		else	{
			// --- track the y axes, slope the x axes
			str = lenx * 2;
			len = leny;
			slope = &begx;
			track = &begy;
		}
		int d = str - len;
		int ang = d - len;
		// --- plot the line
		while (len--)	{
			*track += inctrack;
			if (d <= 0)
				d += str;
			else	{
				*slope += incslope;
				d += ang;
			}
			drawpoint(begx, begy);
		}
	}
}


} // namespace GraphicsLibrary

#endif