#include <TYFC.h>

TYFCApp* TYFCApp::theApp = 0;

TYFCApp::TYFCApp(const char* ttl, int wd, int ht) : 
                 title(ttl), width(wd), height(ht)
{
	theApp = this;
    hdc = 0;
}

inline COLORREF TYFCApp::Windowscolor(Color color) const
{
	static COLORREF crs[] = {
		RGB(0,0,0),
		RGB(0,0,255),
		RGB(0,255,0),
		RGB(0,255,255),
		RGB(255,0,0),
		RGB(255,0,255),
		RGB(255,255,0),
		RGB(255,255,255)
	};
	return crs[color];
}

void TYFCApp::drawpoint(int x, int y, Color color)
{
    SetPixel(hdc, adjx(x), adjy(y), Windowscolor(color));
}

void TYFCApp::drawshapes()
{
    std::list<Shape*>::iterator iter;
    for (iter = shapes.begin(); iter != shapes.end(); iter++)	{
		HPEN pen = CreatePen(PS_SOLID, 1, Windowscolor((*iter)->color));
		HPEN oldpen = SelectObject(hdc, pen);

		LOGBRUSH lb = { BS_HOLLOW };
		HBRUSH brush = CreateBrushIndirect(&lb);
		HBRUSH oldbrush = SelectObject(hdc, brush);

        (*iter)->draw();

		SelectObject(hdc, oldbrush);
		DeleteObject(brush);
		SelectObject(hdc, oldpen);
		DeleteObject(pen);
	}
}
void TYFCApp::addshape(Shape& shape)
{
    shapes.push_back(&shape);
}

extern "C"	{

LRESULT CALLBACK WndProc (HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	if (TYFCApp::theApp == 0)
		return 0;

	switch (msg)   {
		case WM_CREATE:
		case WM_SIZE:
		{
			RECT rc;
			GetClientRect(hwnd, &rc);   			
			TYFCApp::theApp->width = rc.right;
			TYFCApp::theApp->height = rc.bottom;
			if (msg == WM_CREATE)
				TYFCApp::theApp->Initialize();
			return 0 ;
		}
		case WM_PAINT:
		{
			InvalidateRect(hwnd, 0, true);
			TYFCApp::theApp->hdc = BeginPaint(hwnd, &TYFCApp::theApp->ps) ;
			// iterate the shapes and draw them
			TYFCApp::theApp->drawshapes();
			EndPaint(hwnd, &TYFCApp::theApp->ps);
			return 0;
		}
		case WM_DESTROY:
			PostQuitMessage(0);
			return 0 ;
	}
	return DefWindowProc(hwnd, msg, wparam, lparam) ;
}
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    PSTR szCmdLine, int iCmdShow)
{
   if (TYFCApp::theApp == 0)
       return 0;

    WNDCLASSEX wndclass  = {
         sizeof(wndclass),
         CS_HREDRAW | CS_VREDRAW,
         WndProc,
         0, 0,
         hInstance,
         LoadIcon(0, IDI_APPLICATION),
         LoadCursor(0, IDC_ARROW),
         (HBRUSH)GetStockObject(WHITE_BRUSH),
         0,
         "TYFCAppWindow",
         LoadIcon(0, IDI_APPLICATION)
    };
    RegisterClassEx(&wndclass) ;
    TYFCApp::theApp->hwnd = CreateWindow(wndclass.lpszClassName, 
                          TYFCApp::theApp->title,
                          WS_OVERLAPPEDWINDOW,
                          CW_USEDEFAULT, CW_USEDEFAULT,
                          TYFCApp::theApp->width,
                          TYFCApp::theApp->height,
                          0, 0, hInstance, 0) ;

    ShowWindow(TYFCApp::theApp->hwnd, iCmdShow);
    UpdateWindow(TYFCApp::theApp->hwnd);

    MSG msg;
    while (GetMessage (&msg, 0, 0, 0))  {
        TranslateMessage(&msg) ;
        DispatchMessage(&msg) ;
    }
    return msg.wParam ;
}

}        // extern "C"

