/*
 * Author: Andrew D. Fernandes <mailto:andrew@cryptonym.com>
 *         Cryptonym Corporation <http://www.cryptonym.com>
 */

#include "affine.h"


apoint aP;


static apoint smtmp;
static MPI num, den, frac;


void init_apoint() {
  aP = apoint_alloc();
  mpi_set_ui(aP.x,2);
  mpi_fromstr(aP.y,
     "0xF725F621D06D3514127D47C4D97C49534D633D14E817E51AC6AC2D6E856F00C");
  num = mpi_alloc(0);
  den = mpi_alloc(0);
  frac = mpi_alloc(0);
  smtmp = apoint_alloc();
}


void fini_apoint() {
  apoint_free(aP);
  mpi_free(num);
  mpi_free(den);
  mpi_free(frac);
  apoint_free(smtmp);
}


apoint apoint_alloc() {
  apoint p;
  p.x = mpi_alloc(0);
  p.y = mpi_alloc(0);
  return p;
}


void apoint_free(apoint p) {
  mpi_free(p.x);
  mpi_free(p.y);
}


void apoint_double(apoint y, apoint x) {
  if ( mpi_cmp_ui(x.x,0) == 0 && mpi_cmp_ui(x.y,0) == 0 ) {
    mpi_set_ui(y.x,0); mpi_set_ui(y.y,0);
  } else {
    mpi_mul_2exp(den,x.y,1);
    if ( mpi_cmp(den,prime) >= 0 ) mpi_sub(den,den,prime);
    mpi_invm(den,den,prime);
    mpi_mul(num,x.x,x.x);
    mpi_special_mod(num);
    mpi_mul_ui(num,num,3);
    if ( mpi_cmp(num,prime) >= 0 ) mpi_sub(num,num,prime);
    mpi_add(num,num,A);
    if ( mpi_cmp(num,prime) >= 0 ) mpi_sub(num,num,prime);
    mpi_mul(frac,num,den);
    mpi_special_mod(frac);      
    mpi_mul(y.x,frac,frac);
    mpi_special_mod(y.x);
    mpi_sub(y.x,y.x,x.x);
    if ( mpi_cmp_ui(y.x,0) < 0 ) mpi_add(y.x,y.x,prime);
    mpi_sub(y.x,y.x,x.x);
    if ( mpi_cmp_ui(y.x,0) < 0 ) mpi_add(y.x,y.x,prime);
    mpi_sub(y.y,x.x,y.x);
    if ( mpi_cmp_ui(y.y,0) < 0 ) mpi_add(y.y,y.y,prime);
    mpi_mul(y.y,y.y,frac);
    mpi_special_mod(y.y);
    mpi_sub(y.y,y.y,x.y);
    if ( mpi_cmp_ui(y.y,0) < 0 ) mpi_add(y.y,y.y,prime);
  }
}


void apoint_add(apoint z, apoint x, apoint y) {
  if ( mpi_cmp_ui(x.x,0) == 0 && mpi_cmp_ui(x.y,0) == 0 ) {
    mpi_set(z.x,y.x); mpi_set(z.y,y.y);
  } else if ( mpi_cmp_ui(y.x,0) == 0 && mpi_cmp_ui(y.y,0) == 0 ) {
    mpi_set(z.x,x.x); mpi_set(z.y,x.y);
  } else {
    if ( mpi_cmp(x.x,y.x) != 0 ) {
      mpi_sub(den,x.x,y.x);
      if ( mpi_cmp_ui(den,0) < 0 ) mpi_add(den,den,prime);
      mpi_invm(den,den,prime);
      mpi_sub(num,x.y,y.y);
      if ( mpi_cmp_ui(num,0) < 0 ) mpi_add(num,num,prime);
    } else {
      if ( (mpi_cmp(x.y,y.y) != 0) | (mpi_cmp_ui(y.y,0) == 0) ) {
	mpi_set_ui(z.x,0);
	mpi_set_ui(z.y,0);
	return;
      }
      mpi_mul_2exp(den,y.y,1);
      if ( mpi_cmp(den,prime) >= 0 ) mpi_sub(den,den,prime);
      mpi_invm(den,den,prime);
      mpi_mul(num,x.x,x.x);
      mpi_special_mod(num);
      mpi_mul_ui(num,num,3);
      if ( mpi_cmp(num,prime) >= 0 ) mpi_sub(num,num,prime);
      mpi_add(num,num,A);
      if ( mpi_cmp(num,prime) >= 0 ) mpi_sub(num,num,prime);
    }
    mpi_mul(frac,num,den);
    mpi_special_mod(frac);      
    mpi_mul(z.x,frac,frac);
    mpi_special_mod(z.x);
    mpi_sub(z.x,z.x,x.x);
    if ( mpi_cmp_ui(z.x,0) < 0 ) mpi_add(z.x,z.x,prime);
    mpi_sub(z.x,z.x,y.x);
    if ( mpi_cmp_ui(z.x,0) < 0 ) mpi_add(z.x,z.x,prime);
    mpi_sub(z.y,y.x,z.x);
    if ( mpi_cmp_ui(z.y,0) < 0 ) mpi_add(z.y,z.y,prime);
    mpi_mul(z.y,z.y,frac);
    mpi_special_mod(z.y);
    mpi_sub(z.y,z.y,y.y);
    if ( mpi_cmp_ui(z.y,0) < 0 ) mpi_add(z.y,z.y,prime);
  }
}


void apoint_scalar_multiply(apoint ge, apoint g, MPI e) {
  int i;
  struct gcry_mpi tmp;
  unsigned int j;
  mpi_set_ui(ge.x,0);
  mpi_set_ui(ge.y,0);
  for ( i=e->nlimbs-1; i>=0; i-- ) {
    for ( j=(1<<(BITS_PER_MPI_LIMB-1)); j>0; j/=2 ) {
      apoint_double(smtmp,ge);
      tmp = *ge.x; *ge.x = *smtmp.x; *smtmp.x = tmp;
      tmp = *ge.y; *ge.y = *smtmp.y; *smtmp.y = tmp;
      if ( e->d[i] & j ) {
	apoint_add(smtmp,ge,g);
	tmp = *ge.x; *ge.x = *smtmp.x; *smtmp.x = tmp;
	tmp = *ge.y; *ge.y = *smtmp.y; *smtmp.y = tmp;
      }
    }
  }
}
