/* A combined linear congruential pseudo-random number generator from
   "Applied Cryptography" by Bruce Schneier, page 371.  Period is
   (2^31 - 85)(2^31 - 249) .
*/

static long S1 = 1;
static long S2 = 1;

void init_lcg(long s1, long s2)
{
  S1 = s1;
  S2 = s2;
}

double combined_lcg(void)
{
  long q;
  long z;
  
  q = S1 / 53668;
  S1 = 40014 * (S1 - 53668 * q) - 12211 * q;
  if(S1 < 0) {
    S1 += 2147483563L;
  }
  q = S2 / 52774;
  S2 = 40692 * (S2 - 52774 * q) - 3791 * q;
  if(S2 < 0) {
    S2 += 2147483399L;
  }
  if((z = S1 - S2) < 1.0) {
    z += 2147483562L;
  }
  return z * 4.656613e-10;
}
