/* FILE        : hash.c
** AUTHOR      : Jim Shapiro
** DATE        : 19 March, 1999
** DISCLAIMER  : No liability is assumed by the author for any use
**               made of this program.
** DISTRIBUTION: Any use may be made of this program, as long as the
**               clear acknowledgment is made to the author in code
**               and runtime executables.  The author retains
**               copyright.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "hash.h"

static BOOLEAN cb_key_search(int h, char *key);
static unsigned hash(int prime, char *key);

void initialize_cb_hash_count(int count)
{
  if((Cb_sizes = (int *)malloc(count * sizeof(int))) == NULL) {
    (void)printf("Cannot malloc Cb_sizes!...");
    exit(EXIT_FAILURE);
  }
}

void initialize_cb_hash_size(int h, int size)
{
  Cb_sizes[h] = size;
}

void initialize_cb_hashes(int count)
{
  int i;
  int k;
  
  Cb_tail.next = &Cb_tail;
  Cb_tail.value = -1; /* only valid BOOLEANs are FALSE and TRUE */

  if((Cb_heads_ppp = (CB_NODE ***)malloc(count *
                              sizeof(CB_NODE **))) == NULL) {
    (void)puts("Cannot malloc Cb_heads_ppp!...");
    exit(EXIT_FAILURE);
  }
  
  for(k = 0; k < count; k++) {
    int a_size = Cb_sizes[k];
    if((Cb_heads_ppp[k] = (CB_NODE **)malloc(a_size *
                             sizeof(CB_NODE *))) == NULL) {
      (void)printf("Cannot malloc Cb_heads_ppp[%d]!...\n", k);
      exit(EXIT_FAILURE);
    }
    for(i = 0; i < a_size; i++) {
      if((Cb_heads_ppp[k][i] = (CB_NODE *)malloc(sizeof(CB_NODE)))
         == NULL) {
        (void)printf("Cannot malloc Cb_heads_ppp[%d][%d]!...\n",
                     k, i);
        exit(EXIT_FAILURE);
      }
      Cb_heads_ppp[k][i]->next = &Cb_tail;
    }
  }
}

INSERT_STATUS cb_new_pair(int h, char *key, BOOLEAN value)
{
  CB_NODE *x_p;
  CB_NODE *t_p = Cb_heads_ppp[h][hash(Cb_sizes[h], key)];
  INSERT_STATUS rv = DUPLICATE_KEY;

  if((Cb_tail.key = strdup(key)) == NULL) {
    (void)puts("Cannot strdup key to Cb_tail.key!...");
    exit(EXIT_FAILURE);
  }
  while(strcmp(key, t_p->next->key) > 0) {
    t_p = t_p->next;
  }
  
  /* Only insert if we encountered a key greater than this one or we
     ended up going all the way to the tail without finding this key.
  */
  if((strcmp(key, t_p->next->key) != 0) ||
     (t_p->next->next == t_p->next)) {
    if((x_p = (CB_NODE *)malloc(sizeof(CB_NODE))) == NULL) {
      (void)puts("Cannot malloc x_p for new CB_NODE!...");
      exit(EXIT_FAILURE);
    }
    rv = INSERT_OK;
    x_p->next = t_p->next;
    t_p->next = x_p;
    if((x_p->key = strdup(key)) == NULL) {
      (void)puts("Cannot strdup key to x_p.key!...");
      exit(EXIT_FAILURE);
    }
    x_p->value = value;
  }
  free(Cb_tail.key);
  return rv;
}

BOOLEAN cb_exists(int h, char *key)
{
  return (cb_key_search(h, key) == -1) ? FALSE : TRUE;
}

static BOOLEAN cb_key_search(int h, char *key)
{
  int rv = Cb_tail.value;
  /* Start at first link */
  CB_NODE *t_p = Cb_heads_ppp[h][hash(Cb_sizes[h], key)]->next;
  
  if((Cb_tail.key = strdup(key)) == NULL) {
    (void)puts("Cannot strdup key to Cb_tail.key!...");
    exit(EXIT_FAILURE);
  }
  while(strcmp(key, t_p->key) > 0) {
    t_p = t_p->next;
  }
  if(strcmp(key, t_p->key) == 0) {
    rv = t_p->value;
  }
  free(Cb_tail.key);
  return rv;
}

static unsigned hash(int prime, char *key)
{
  int h;
  for(h = 0; *key != '\0'; key++) {
    h = ((h << 6) + *key) % prime;
  }
  return h;
}
