/* FILE        : hmac.c
** AUTHOR      : Jim Shapiro
** DATE        : 19 March, 1999
** DISCLAIMER  : No liability is assumed by the author for any use
**               made of this program.
** DISTRIBUTION: Any use may be made of this program, as long as the
**               clear acknowledgment is made to the author in code
**               and runtime executables.  The author retains
**               copyright.
*/

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>  /* bcopy() and bzero() */

#include "msg_dgst_rsa.h"

void hmac_md5_string(unsigned char *text_p, unsigned int text_len,
                     unsigned char *key_p , unsigned int key_len,
                     unsigned char *digest_p)
{ /* digest_p will be 16 bytes */
  unsigned char *ipad_p;
  unsigned char opad_p[80]; /* 64 for opad and 16 for MD5 */
  int i;

  if(key_len > 64) {
    unsigned char tmp_key_p[16];
    
    md5_string(key_p, key_len, tmp_key_p); /* key = MD5(key) */
    bcopy(tmp_key_p, key_p, 16);
    key_len = 16;
  }
  if((ipad_p = (unsigned char *)calloc((64 + text_len),
                                  sizeof(unsigned char))) == NULL) {
    (void)fprintf(stderr, "Cannot calloc ipad_p at %u bytes!...\n",
                  64 + text_len);
    exit(EXIT_FAILURE);
  }
  /* 0 key padding done with calloc() */
  bcopy(key_p , ipad_p     , key_len);
  bcopy(text_p, ipad_p + 64, text_len);

  bzero(opad_p, 64);
  bcopy(key_p, opad_p, key_len);
  for(i = 0; i < 64; i++) {
    ipad_p[i] ^= 0x36;
    opad_p[i] ^= 0x5c;
  }
  md5_string(ipad_p, 64 + text_len, opad_p + 64);
  md5_string(opad_p, 80, digest_p);
  free(ipad_p);
}

void hmac_sha_string(unsigned char *text_p, unsigned int text_len,
                     unsigned char *key_p , unsigned int key_len,
                     unsigned char *digest_p)
{ /* digest_p will be 20 bytes */
  unsigned char *ipad_p;
  unsigned char opad_p[84]; /* 64 for opad and 20 for SHA */
  int i;

  if(key_len > 64) {
    unsigned char tmp_key_p[20];

    sha_string(key_p, key_len, tmp_key_p); /* key = SHA(key) */
    bcopy(tmp_key_p, key_p, 20);
    key_len = 20;
  }
  
  if((ipad_p = (unsigned char *)calloc((64 + text_len),
                                  sizeof(unsigned char))) == NULL) {
    (void)fprintf(stderr, "Cannot calloc ipad_p at %u bytes!...\n",
                  64 + text_len);
    exit(EXIT_FAILURE);
  }
  /* 0 key padding done with calloc() */
  bcopy(key_p , ipad_p     , key_len);
  bcopy(text_p, ipad_p + 64, text_len);

  bzero(opad_p, 64);
  bcopy(key_p, opad_p, key_len);
  for(i = 0; i < 64; i++) {
    ipad_p[i] ^= 0x36;
    opad_p[i] ^= 0x5c;
  }
  sha_string(ipad_p, 64 + text_len, opad_p + 64);
  sha_string(opad_p, 84, digest_p);
  free(ipad_p);
}
