/* FILE        : package.c
** AUTHOR      : Jim Shapiro
** DATE        : 19 March, 1999
** DISCLAIMER  : No liability is assumed by the author for any use
**               made of this program.
** DISTRIBUTION: Any use may be made of this program, as long as the
**               clear acknowledgment is made to the author in code
**               and runtime executables.  The author retains
**               copyright.
*/

#ifdef THIS_IS_MAIN
#define GLOBAL
#else
#define GLOBAL extern
#endif

#include <stdlib.h>
#include <stdio.h>
#include <strings.h> /* bzero() */

#include "blowfish.h"
#include "package.h"
#include "clcg_rand.h"

#define KEY_LEN 8

/* public key */
static unsigned char Pub_key_p[] = { 0x7C, 0xA1, 0x10, 0x45,
                                     0x4A, 0x1A, 0x6E, 0x57 };
static ARRAYS P_array; /* for the public key */

typedef union
{
  unsigned char c[8];
  unsigned long data_p[2];
} B;

void package_one_time_set(long s1, long s2)
{ /* call once before using any package functions */
  init_lcg(s1, s2);
  blowfish_initialize(&P_array, Pub_key_p, KEY_LEN);
}

void package_initialize(ARRAYS *r_array_p, unsigned char *mp_s)
{ /* call once to set random key and "random" Blowfish array */
  unsigned char ran_key_p[KEY_LEN]; /* random key */
  unsigned int i;
  
  /* initialize last output block */
  for(i = 0; i < KEY_LEN; i++) {
    mp_s[i] = ran_key_p[i] = (unsigned char)(256 * combined_lcg());
  }
  blowfish_initialize(r_array_p, ran_key_p, KEY_LEN);
}

void package_transform(unsigned char *mp_s, ARRAYS *r_array_p,
                       unsigned long i, unsigned char m[8],
                       unsigned char mp[8])
{ /* call s times, once for each block m -- sets pseudo-block mp
     and updates last block */
  int j;
  B b;

  b.data_p[0] = i;
  b.data_p[1] = 0l;
  blowfish_encrypt(r_array_p, b.data_p, 1); /* E(K', i) */
  for(j = 0; j < KEY_LEN; j++) {            /* m xor E(K', i) */
    mp[j] = m[j] ^ b.c[j];
  }
  b.data_p[0] = i;
  b.data_p[1] = 0l;
  for(j = 0; j < KEY_LEN; j++) {            /* mi' xor i */
    b.c[j] ^= mp[j];
  }
  /* hi = E(K0, mi' xor i) */
  blowfish_encrypt(&P_array, b.data_p, 1);
  for(j = 0; j < KEY_LEN; j++) { /* K' xor h0 xor ... xor hi */
    mp_s[j] ^= b.c[j];
  }
}

void rev_package_initialize(unsigned char *mp_s)
{
  bzero(mp_s, 8); /* zero out for xor */
}

void rev_package_build_kp(unsigned char *mp_s, unsigned long i,
                          unsigned char mp[8])
{ /* call s times, after which mp_s should be ran_key_p */
  int j;
  B b;
  
  b.data_p[0] = i;
  b.data_p[1] = 0l;
  for(j = 0; j < KEY_LEN; j++) {           /* mi' xor i */
    b.c[j] ^= mp[j];
  }
  /* hi = E(K0, mi' xor i) */
  blowfish_encrypt(&P_array, b.data_p, 1);
  for(j = 0; j < KEY_LEN; j++) { /* ms' xor h0 xor ... xor hi */
    mp_s[j] ^= b.c[j];
  }
}

void rev_package_set_key(unsigned char *mp_s, unsigned char *hold_p,
                         ARRAYS *r_array_p)
{ /* call once after rev_package_build_kp is called for the last
     (s'th) time */
  int i;
  
  for(i = 0; i < 8; i++) {
    mp_s[i] ^= hold_p[i];
  }
  blowfish_initialize(r_array_p, mp_s, KEY_LEN);
}

void rev_package_finalize(ARRAYS *r_array_p, unsigned long i,
                          unsigned char mp[8], unsigned char m[8])
{ /* call once for each of the s blocks (in any order) */
  int j;
  B b;

  b.data_p[0] = i;
  b.data_p[1] = 0l;
  blowfish_encrypt(r_array_p, b.data_p, 1); /* E(K', i) */
  for(j = 0; j < KEY_LEN; j++) { /* mi = mi' xor E(K', i) */
    m[j] = mp[j] ^ b.c[j];
  }
}
