/*
 * Author: Andrew D. Fernandes <mailto:andrew@cryptonym.com>
 *         Cryptonym Corporation <http://www.cryptonym.com>
 */

#include "projective.h"


ppoint pP;
MPI A, B, n;


static ppoint smtmp;
static MPI t1, t2, t3, t4, t5, t6, t7;


void init_ppoint() {
  pP = ppoint_alloc();
  mpi_set_ui(pP.x,2);
  mpi_fromstr(pP.y,
     "0xF725F621D06D3514127D47C4D97C49534D633D14E817E51AC6AC2D6E856F00C");
  mpi_set_ui(pP.z,1);
  smtmp = ppoint_alloc();
  t1 = mpi_alloc(0);
  t2 = mpi_alloc(0);
  t3 = mpi_alloc(0);
  t4 = mpi_alloc(0);
  t5 = mpi_alloc(0);
  t6 = mpi_alloc(0);
  t7 = mpi_alloc(0);
}


void fini_ppoint() {
  ppoint_free(pP);
  ppoint_free(smtmp);
  mpi_free(t1);
  mpi_free(t2);
  mpi_free(t3);
  mpi_free(t4);
  mpi_free(t5);
  mpi_free(t6);
  mpi_free(t7);
}


ppoint ppoint_alloc() {
  ppoint p;
  p.x = mpi_alloc(0);
  p.y = mpi_alloc(0);
  p.z = mpi_alloc(0);
  return p;
}


void ppoint_free(ppoint p) {
  mpi_free(p.x);
  mpi_free(p.y);
  mpi_free(p.z);
}


void ppoint_double(ppoint y, ppoint x) {
  if ( ( mpi_cmp_ui(x.y,0) == 0 ) || ( mpi_cmp_ui(x.z,0) == 0 ) ) {
    mpi_set_ui(y.x,1);
    mpi_set_ui(y.y,1);
    mpi_set_ui(y.z,0);
    return;
  }
  mpi_set(t1,x.x);
  mpi_set(t2,x.y);
  mpi_set(t3,x.z);
  mpi_mul(t4,t3,t3);
  mpi_special_mod(t4);
  mpi_sub(t5,t1,t4);
  if ( mpi_cmp_ui(t5,0) < 0 ) mpi_add(t5,t5,prime);
  mpi_add(t4,t1,t4);
  if ( mpi_cmp(t4,prime) >= 0 ) mpi_sub(t4,t4,prime);
  mpi_mul(t5,t4,t5);
  mpi_special_mod(t5);
  mpi_mul_ui(t4,t5,3);
  mpi_special_mod(t4);
  mpi_mul(t3,t2,t3);
  mpi_special_mod(t3);
  mpi_mul_2exp(t3,t3,1);
  mpi_special_mod(t3);
  mpi_mul(t2,t2,t2);
  mpi_special_mod(t2);
  mpi_mul(t5,t1,t2);
  mpi_special_mod(t5);
  mpi_mul_2exp(t5,t5,2);
  mpi_special_mod(t5);
  mpi_mul(t1,t4,t4);
  mpi_special_mod(t1);
  mpi_sub(t1,t1,t5);
  if ( mpi_cmp_ui(t1,0) < 0 ) mpi_add(t1,t1,prime);
  mpi_sub(t1,t1,t5);
  if ( mpi_cmp_ui(t1,0) < 0 ) mpi_add(t1,t1,prime);
  mpi_mul(t2,t2,t2);
  mpi_special_mod(t2);
  mpi_mul_2exp(t2,t2,3);
  mpi_special_mod(t2);
  mpi_sub(t5,t5,t1);
  if ( mpi_cmp_ui(t5,0) < 0 ) mpi_add(t5,t5,prime);
  mpi_mul(t5,t4,t5);
  mpi_special_mod(t5);
  mpi_sub(t2,t5,t2);
  if ( mpi_cmp_ui(t2,0) < 0 ) mpi_add(t2,t2,prime);
  mpi_set(y.x,t1);
  mpi_set(y.y,t2);
  mpi_set(y.z,t3);
}


void ppoint_add(ppoint z, ppoint x, ppoint y) {
  /* warning: does not properly add the point at infinty */
  mpi_set(t1,x.x);
  mpi_set(t2,x.y);
  mpi_set(t3,x.z);
  mpi_set(t4,y.x);
  mpi_set(t5,y.y);
  if ( mpi_cmp_ui(y.z,1) != 0 ) {
    mpi_set(t6,y.z);
    mpi_mul(t7,t6,t6);
    mpi_special_mod(t7);
    mpi_mul(t1,t1,t7);
    mpi_special_mod(t1);
    mpi_mul(t7,t6,t7);
    mpi_special_mod(t7);
    mpi_mul(t2,t2,t7);
    mpi_special_mod(t2);
  }
  mpi_mul(t7,t3,t3);
  mpi_special_mod(t7);
  mpi_mul(t4,t4,t7);
  mpi_special_mod(t4);
  mpi_mul(t7,t3,t7);
  mpi_special_mod(t7);
  mpi_mul(t5,t5,t7);
  mpi_special_mod(t5);
  mpi_sub(t4,t1,t4);
  if ( mpi_cmp_ui(t4,0) < 0 ) mpi_add(t4,t4,prime);
  mpi_sub(t5,t2,t5);
  if ( mpi_cmp_ui(t5,0) < 0 ) mpi_add(t5,t5,prime);
  if ( mpi_cmp_ui(t4,0) == 0 ) {
    if ( mpi_cmp_ui(t5,0) == 0 ) {
      ppoint_double(z,x);
      return;
    } else {
      mpi_set_ui(z.x,1);
      mpi_set_ui(z.y,1);
      mpi_set_ui(z.z,0);
      return;
    }
  }
  mpi_mul_2exp(t1,t1,1);
  if ( mpi_cmp(t1,prime) >=0 ) mpi_sub(t1,t1,prime);
  mpi_sub(t1,t1,t4);
  if ( mpi_cmp_ui(t1,0) < 0 ) mpi_add(t1,t1,prime);
  mpi_mul_2exp(t2,t2,1);
  if ( mpi_cmp(t2,prime) >=0 ) mpi_sub(t2,t2,prime);
  mpi_sub(t2,t2,t5);
  if ( mpi_cmp_ui(t2,0) < 0 ) mpi_add(t2,t2,prime);
  if ( mpi_cmp_ui(y.z,1) != 0 ) {
    mpi_mul(t3,t3,t6);
    mpi_special_mod(t3);
  }
  mpi_mul(t3,t3,t4);
  mpi_special_mod(t3);
  mpi_mul(t7,t4,t4);
  mpi_special_mod(t7);
  mpi_mul(t4,t4,t7);
  mpi_special_mod(t4);
  mpi_mul(t7,t1,t7);
  mpi_special_mod(t7);
  mpi_mul(t1,t5,t5);
  mpi_special_mod(t1);
  mpi_sub(t1,t1,t7);
  if ( mpi_cmp_ui(t1,0) < 0 ) mpi_add(t1,t1,prime);
  mpi_sub(t7,t7,t1);
  if ( mpi_cmp_ui(t7,0) < 0 ) mpi_add(t7,t7,prime);
  mpi_sub(t7,t7,t1);
  if ( mpi_cmp_ui(t7,0) < 0 ) mpi_add(t7,t7,prime);
  mpi_mul(t5,t5,t7);
  mpi_special_mod(t5);
  mpi_mul(t4,t2,t4);
  mpi_special_mod(t4);
  mpi_sub(t2,t5,t4);
  if ( mpi_cmp_ui(t2,0) < 0 ) mpi_add(t2,t2,prime);
  if ( t2->d[0] & 1 ) mpi_add(t2,t2,prime);
  mpi_rshift(t2,t2,1);
  mpi_set(z.x,t1);
  mpi_set(z.y,t2);
  mpi_set(z.z,t3);
}


void normalize_ppoint(ppoint p) {
  /* should never be the point at infinity */
  mpi_invm(p.z,p.z,prime);
  mpi_mul(p.y,p.z,p.y);
  mpi_special_mod(p.y);
  mpi_mul(p.z,p.z,p.z);
  mpi_special_mod(p.z);
  mpi_mul(p.x,p.z,p.x);
  mpi_special_mod(p.x);
  mpi_mul(p.y,p.z,p.y);
  mpi_special_mod(p.y);
  mpi_set_ui(p.z,1);
}


void ppoint_scalar_multiply(ppoint ge, ppoint g, MPI e) {
  int i, flag = 0;
  struct gcry_mpi tmp;
  unsigned int j;
  for ( i=e->nlimbs-1; i>=0; i-- ) {
    for ( j=(1<<(BITS_PER_MPI_LIMB-1)); j>0; j/=2 ) {
      if ( flag ) {
	ppoint_double(smtmp,ge);
	tmp = *ge.x; *ge.x = *smtmp.x; *smtmp.x = tmp;
	tmp = *ge.y; *ge.y = *smtmp.y; *smtmp.y = tmp;
	tmp = *ge.z; *ge.z = *smtmp.z; *smtmp.z = tmp;
      }
      if ( e->d[i] & j ) {
	if ( flag ) {
	  ppoint_add(smtmp,ge,g);
	  tmp = *ge.x; *ge.x = *smtmp.x; *smtmp.x = tmp;
	  tmp = *ge.y; *ge.y = *smtmp.y; *smtmp.y = tmp;
	  tmp = *ge.z; *ge.z = *smtmp.z; *smtmp.z = tmp;
	} else {
	  mpi_set(ge.x,g.x);
	  mpi_set(ge.y,g.y);
	  mpi_set(ge.z,g.z);
	  flag = 1;
	}
      }
    }
  }
}
